def Group(ctx):
    return _GroupUngroup(ctx, "Group")
def Ungroup(ctx):
    return _GroupUngroup(ctx, "Ungroup")

def _GroupUngroup(ctx, _cmd="Group"):
    _hosts = []
    _hostgroups = []
    _templates = []
    while True:
        p = ctx.pull()
        if not p:
            break
        if p.has_key("HOST"):
            _hosts += extract_key_from_info(p,"HOST","hostid")
        elif p.has_key("HOSTGROUPS"):
            _hostgroups += extract_key_from_info(p,"HOSTGROUPS","groupid")
        elif p.has_key("TEMPLATES"):
            _templates += extract_key_from_info(p,"TEMPLATES","templateid")
        else:
            ctx.push(p)
            break
    if len(_hostgroups) == 0:
        Status(ctx, STATUS=False, MESSAGE="There is no (Hostgroups) for the (Group)")
        return ctx
    if len(_hosts) == 0 and len(_templates) == 0:
        Status(ctx, STATUS=False, MESSAGE="There is no (Hosts) nor (Templates) for the (Group)")
        return ctx
    for i in _hostgroups:
        _hi = ctx.zapi.hostgroup.get(groupids=[i,],selectHosts=1,selectTemplates=1)
        if len(_hi) == 0:
            Status(ctx, STATUS=False, MESSAGE="For some reason can not find (Hostgroup (id)-%s)"%i)
            return ctx
        else:
            _hi = _hi[0]
        _hi["hosts"] = listofdict2list(_hi["hosts"], "hostid")
        _hi["templates"] = listofdict2list(_hi["templates"], "templateid")
        for j in _hosts:
            if ctx.env.shell != None:
                ctx.env.shell.ok("(%s (id): %s and (Host (id): %s))"%(_cmd,i,j))
            if _cmd == "Group":
                if j not in _hi["hosts"]:
                    _hi["hosts"].append(j)
            elif _cmd == "Ungroup":
                if j in _hi["hosts"]:
                    _hi["hosts"].remove(j)
            else:
                pass
        for k in _templates:
            if k not in _hi["templates"]:
                if ctx.env.shell != None:
                    ctx.env.shell.ok("(%s (id): %s and (Template (id): %s))"%(_cmd,i,k))
                if _cmd == "Group":
                    _hi["templates"].append(k)
                elif _cmd == "Ungroup":
                    _hi["templates"].remove(k)
                else:
                    pass
        ctx.zapi.hostgroup.massupdate(groups=[{"groupid":i},],
                                      hosts=list2listofdicts(_hi["hosts"],"hostid"),
                                      templates=list2listofdicts(_hi["templates"],"templateid"))

    Status(ctx, STATUS=True, MESSAGE="(Group) returns TRUE")
    return ctx





def Filter(ctx, _comparator=None, *_arg_filter, **_kw_filter):
    data = ctx.pull()
    if not data:
        return ctx
    if type(data) != type({}):
        ctx.push(data)
        return ctx
    if ctx.env.shell != None:
        for _p in _arg_filter:
            ctx.env.shell.ok("[Filter] %s"%repr(_p))
        for _p in _kw_filter.keys():
            ctx.env.shell.ok("{Filter} %s->%s->%s" % (_p, repr(_comparator), repr(_kw_filter[_p])))
    new_data = {}
    for k in data.keys():
        new_data[k] = []
    filters = []
    for _p in _arg_filter:
        if len(_p) == 2 and _comparator not in [None, False]:
            _key, _f = _p
            filters.append({'KEY': _key, 'C': _comparator, 'BASE': _f})
        elif len(_p) == 3:
            _key, _c, _f = _p
            filters.append({'KEY': _key, 'C': _c, 'BASE': _f})
        else:
            continue
    if _comparator != None:
        for _key, _f in _kw_filter.items():
            filters.append({'KEY': _key, 'C': _comparator, 'BASE': _f})
    for k in data.keys():
        for v in data[k]:
            succ = False
            for f in filters:
                if not v.has_key(f['KEY']):
                    continue
                if type(f['BASE']) in [types.LongType, types.IntType]:
                    x = long(v[f['KEY']])
                elif type(f['BASE']) == types.FloatType:
                    x = float(v[f['KEY']])
                else:
                    x = v[f['KEY']]
                if f['C'](x, f['BASE']):
                    succ = True
                else:
                    succ = False
            if succ:
                new_data[k].append(v)
    ctx.push(new_data)
    return ctx


import argparse
import os
import sys

class ZQ_GEN(object):
    def __init__(self, _desc, _epilog):
        self.epilog = _epilog + ". Type: %s help for the general help"%sys.argv[0]
        self.BANNER="GENERIC"
        try:
            self.ZQ_HOME = get_from_env("ZQ_HOME", default="%s/.zq"%os.environ["HOME"])
        except:
            self.ZQ_HOME = "/tmp"
        self.parser = argparse.ArgumentParser(prog='zql', description=_desc, epilog=self.epilog)
        self.parser.add_argument("--banner", action="store_true", help="Display banner during start")
        self.parser.add_argument("--home", "-H", type=str, default=self.ZQ_HOME,
                                 help="Path to the default home directory or URL")
        self.parser.add_argument('N', metavar='N', type=str, nargs='*',
                                 help='Parameters')
        self.ready = True
    def _mkbanner(self):
        return str_dict(self.env.cfg)
    def banner(self):
        b = color(banner(self.BANNER), "magenta")
        b += self._mkbanner()
        print b
        self.important("VERSION %s"%ZQ_VERSION)
    def _main_preflight(self):
        self._call_hiera("preflight", "Preflight check in %s is failed")
    def _call_hiera(self, name, err_msg):
        visited = []
        for b in self.__class__.__bases__:
            if b.__name__ in visited:
                continue
            else:
                visited.append(b.__name__)
            try:
                m = getattr(b, name)
            except AttributeError:
                continue
            if m != None:
                try:
                    self.ok("Calling preflight in %s"%b.__name__)
                    res = apply(m, (self,), {})
                except KeyboardInterrupt:
                    res = False
            if res == False:
                self.error(err_msg%b.__name__)
                sys.exit(98)
    def preflight(self):
        global ENV
        try:
            import humanfriendly
        except ImportError:
            self.important("Module 'humanfriendly' not found. Many parameters will be defaulted")
        if self.env.ready != True:
            self.ready = False
            return False
        self.env.cfg["ZQ_HOME"] = self.args.home
        self.env.cfg["ZQ_ENV_NAME"] = self.args.default_environment
        self.env.cfg["ZQ_ENV_PATH"] = "%s/%s/"%(self.env.cfg["ZQ_HOME"], self.env.cfg["ZQ_ENV_NAME"])
        if self.env.cfg["ZQ_ENV_PATH"] == "+":
            self.env.cfg["ZQ_ENV_PATH"] = "+" + posixpath.abspath(self.env.cfg["ZQ_ENV_PATH"][1:])
        elif self.env.cfg["ZQ_ENV_PATH"] != "@":
            self.env.cfg["ZQ_ENV_PATH"] = "+" + posixpath.abspath(self.env.cfg["ZQ_ENV_PATH"])
        else:
            pass
        if not check_reference_read(self.env.cfg["ZQ_ENV_PATH"], True):
            self.warning("Configuration HOME %s can not be accessed"%self.env.cfg["ZQ_ENV_PATH"])
        else:
            self.ok("Configuration HOME set to %s"%self.env.cfg["ZQ_ENV_PATH"])
        return True
    def process(self):
        self.args = self.parser.parse_args()
        #print self.args
        self._call_hiera("make_doc", "Error creating documentation in %s")
        self.env = ZQ_ENV(self)
        ENVIRONMENT(self.args.default_environment,self.env)
        self.main_preflight()
        if self.args.banner:
            self.banner()
        for e in ENV.keys():
            for s in ENV[e].srv.keys():
                self.ok("ENV: %s Server: %s - %s"%(e,s,ENV[e].srv[s].url))
        if len(self.args.N) == 0:
            self.error("You did not specified the command. Please run %s help"%sys.argv[0])
            self.ready = False
            return False
        if len(self.args.N) > 1 and self.args.N[1].upper() == "HELP":
            cmd = getattr(self, "HELP_"+self.args.N[0].upper(), None)
        else:
            cmd = getattr(self, self.args.N[0].upper(), None)
        if cmd == None:
            self.error("Command %s not found"%self.args.N[0])
            self.ready = False
            return False
        try:
            apply(cmd, (), {})
        except KeyboardInterrupt:
            self.ready = False
        finally:
            self.ready = True
        return True
    def make_doc(self):
        pass

class ZQ_HELP:
    def __init__(self):
        pass
    def make_doc(self):
        self.doc.append(("help", "Get help about comamnds"))
        self.doc.append(("<command> --help/-h", "Get help about particular comamnds"))
        return True
    def HELP_HELP(self):
        print """Receive the general help about commands or the particular command:
        "help" - list of the available commands
        "<command name> help - help abou specific command
        """
    def HELP(self):
        print "/"+"*"*78+"+"
        for d in self.doc:
            print ": %-20s : %-53s :"%d
        print "+"+"*"*78+"+"





from gevent import monkey
monkey.patch_all()
import sys
import UserDict
import time
import simplejson
import urllib
import types
import gevent
import uuid
import string
import posixpath
import copy
import Queue
from hy.lex import parser, lexer
from hy.importer import hy_eval
from hy import HyList


include "zq_symbols.pyx"
include "zq_constants.pyi"
include "zq_lib.pyx"
include "zq_pyzabbix.pyx"
include "zq_cfg.pyx"
include "zq_env.pyx"
include "zq_bjq.pyx"
include "zq_query.pyx"
include "zq_query_display.pyx"
include "zq_query_status.pyx"
include "zq_query_vars.pyx"
include "zq_query_stackless.pyx"
include "zq_query_logic.pyx"
include "zq_query_query.pyx"
include "zq_query_delete.pyx"
include "zq_query_create.pyx"
include "zq_query_update.pyx"
include "zq_query_new.pyx"
include "zq_query_words.pyx"
include "zq_query_group.pyx"
include "zq_query_output.pyx"
include "zq_query_stack.pyx"
include "zq_query_filters.pyx"
include "zq_clear.pyx"






##
## Update functors
##


def Set(_src, _dst, key, value):
    if not _src.has_key(key):
        return _dst
    _dst[key] = value
    return _dst

def Append(_src, _dst, key, value):
    if not _src.has_key(key) or type(_src[key]) != types.ListType:
        return _dst
    elif _dst.has_key(key) and type(_dst[key]) == types.ListType:
        _dst[key].append(value)
    elif _src.has_key(key) and not _dst.has_key(key) and type(_src[key]) == types.ListType:
        _dst[key] = copy.copy(_src[key])
        _dst[key].append(value)
    else:
        return _dst
    return _dst


##
## Update logic
##

def _update_element(ctx, _cmd, args, _primary_key, _p):
    for p in _p:
        if not p.has_key(_primary_key):
            return False
        d = {_primary_key:p[_primary_key]}
        for _values in args:
            if len(_values) == 2:
                _val = None
                key, func = _values
            elif len(_values) == 3:
                key, func, _val = _values
            else:
                ## Bad parameters
                continue
            try:
                d = apply(func, (p, d, key, _val))
            except KeyboardInterrupt:
                return False
        res = apply(_cmd, (), d)
        if not res.has_key(_primary_key) or len(res[_primary_key]) == 0:
            print 1,res
            return False
    return True


def _update_hostgroup(ctx, args, kw, _data):
    print "YYY",_data
    return _update_element(ctx, ctx.zapi.hostgroup.update, args, "groupid", _data)

def _update_host(ctx, args, kw, _data):
    print "(Update) HOST",_data
    return True

_UPDATE_CALL_TABLE={
    "HOSTGROUPS": _update_hostgroup,
    "HOST": _update_host,
}

def Update(ctx, *args, **kw):
    if not _fill_bjq_queue(ctx, ctx.jobs.direct):
        if ctx.env.shell != None:
            ctx.env.shell.warning("(Update...) can not populate the queue. Check logic!")
        return ctx
    while True:
        sjob = ctx.jobs.pull()
        if sjob == None:
            break
        _pri, _req = sjob
        _key, _data = _req
        if not _UPDATE_CALL_TABLE[_key](ctx, args, kw, _data):
            if ctx.env.shell != None:
                ctx.env.shell.warning("(Update...) can not create element %s"%str(_req))
    return ctx


class ZQ_CMD_QUERY:
    def __init__(self):
        self.URL = get_from_env("ZQ_URL", default="http://127.0.0.1/zabbix")
        self.USER = get_from_env("ZQ_USER", default="Admin")
        self.PASS = get_from_env("ZQ_PASS", default="zabbix")
        self.ZQ_CFG = get_from_env("ZQ_CFG", default=None)
        self.SENDER = get_from_env("ZQ_SENDER", default="127.0.0.1")
        self.REFBASE = get_from_env("ZQ_REF_BASE", default="+.")
        try:
            self.MAX_PIPELINE = int(get_from_env("ZQ_MAX_PIPELINE", default="100"))
        except:
            self.MAX_PIPELINE = 100
        try:
            self.MAX_ENVSTACK = int(get_from_env("ZQ_MAX_ENV_STACK", default="100"))
        except:
            self.MAX_ENVSTACK = 100
        try:
            self.SENDER_PORT = int(get_from_env("ZQ_SENDER_PORT", default="10051"))
        except:
            self.MAX_ENVSTACK = 10051
        self.parser.add_argument("--config", "-c", type=str, default=self.ZQ_CFG,
                                 help="Path to the servers configuration file")
        self.parser.add_argument("--url", type=str, default=self.URL,
                                 help="Zabbix server URL")
        self.parser.add_argument("--sender", type=str, default=self.SENDER,
                                 help="IP address for the Zabbix Sender Protocol")
        self.parser.add_argument("--sender-port", type=int, default=self.SENDER_PORT,
                                 help="Maximum number of elements in the query pipeline")
        self.parser.add_argument("--user", "-U", type=str, default=self.USER,
                                 help="Zabbix User")
        self.parser.add_argument("--password", type=str, default=self.PASS,
                                 help="Zabbix password")
        self.parser.add_argument("--name", type=str, default="zabbix",
                                 help="Zabbix name")
        self.parser.add_argument("--max-query-pipeline", type=int, default=self.MAX_PIPELINE,
                                 help="Maximum number of elements in the query pipeline")
        self.parser.add_argument("--max-env-stack", type=int, default=self.MAX_ENVSTACK,
                                 help="Maximum number of elements in the query pipeline")
        self.parser.add_argument("--ref-base", type=str, default=self.REFBASE,
                                 help="Default base for all Zabbix Query load as column-separated list of the references")

    def preflight(self):
        if self.args.config != None:
            self.ok("Attempting to load %s"%self.args.config)
            cfg_file = "%s/%s"%(self.env.cfg["ZQ_ENV_PATH"], self.args.config)
            self.ok("Loading from %s"%cfg_file)
            cfg = load_config_file(cfg_file)
            if cfg == None:
                self.error("Configuration file %s can not be loaded"%self.args.config)
                return False
            for s in cfg:
                self.env.srv.addServer(s["url"],s["username"],s["password"],s["name"],s["sender"],s["sender_port"])
        else:
            self.env.srv.addServer(self.args.url, self.args.user, self.args.password, self.args.name, self.args.sender, self.args.sender_port)
        self.env.cfg["ZQ_MAX_PIPELINE"] = self.args.max_query_pipeline
        self.env.cfg["ZQ_MAX_ENV_STACK"] = self.args.max_env_stack
        self.env.cfg["ZQ_REF_BASE"] = split_list(self.args.ref_base, ":")
        return True

    def make_doc(self):
        self.doc.append(("query", "Send query to Zabbix"))
    def HELP_QUERY(self):
        print """Send the query to the Zabbix Server:
"query" "(QUERY statement)" "(Query Statement)"

If Query statements prepending with '@', it treated as URL which is loaded and processed
If Query statement prepended with '+', it is treated as path to the local filename with Query statement

Example(s):
    %s query "(ZBX) (Hosts)"
    %s query @http://127.0.0.1/myquery.zql
    %s query +/tmp/myquery.zql
"""%(sys.argv[0],sys.argv[0],sys.argv[0])
    def QUERY(self):
        self.Exec(self.env.QUERY, self.args.default_environment, self.args.default_server)



##
##
##
import os
import posixpath
import fnmatch


def rchop(thestring, ending):
    """Chopping a string ending:

    1. thestring: source string
    2. ending: the tail which shall be removed"""
    if thestring.endswith(ending):
        return thestring[:-len(ending)]
    return thestring

def check_file(fname, mode):
    """
    Check if file exists and we can access it.

    1. :param fname: Filename
    2. :param mode: Access mode
    3. :return: True/False
    """
    fname = os.path.expandvars(fname)
    if os.path.exists(fname) and os.path.isfile(fname) and os.access(fname, mode):
        return True
    return False

def check_directory(dname):
    """
    Check if parameter is directory and we do have READ access to it

    1. :param dname: Directory name
    2. :return: True/False
    """
    dname = os.path.expandvars(dname)
    if os.path.exists(dname) and os.path.isdir(dname) and os.access(dname, os.R_OK):
        return True
    return False

def check_directory_write(dname):
    """
    Check if parameter is directory and we do have WRITE access to it

    1. :param dname: Directory name
    2. :return: True/False
    """
    dname = os.path.expandvars(dname)
    if os.path.exists(dname) and os.path.isdir(dname) and os.access(dname, os.W_OK):
        return True
    return False


def check_file_read(fname):
    """
    Check if parameter is a file and we do have READ access to it

    1. :param fname: Filename
    2. :return: True/False
    """
    return check_file(fname, os.R_OK)

def check_module(fname):
    """
    Check if parameter is a file, we do have READ access to it and it's size is greater than '0'

    1. :param fname: Filename
    2. :return: True/False
    """
    if not check_file_read(fname):
        return False
    if os.path.getsize(fname) > 0:
        return True
    return False

def check_file_write(fname):
    """
    Check if parameter is a file and we do have WRITE access to it

    1. :param fname: Filename
    2. :return: True/False
    """
    return check_file(fname, os.W_OK)

def check_file_exec(fname):
    """
    Check if parameter is a file and we do have EXEC access to it

    1. :param fname: Filename
    2. :return: True/False
    """
    return check_file(fname, os.X_OK)

def get_dir_content(dname):
    """
    Returning the list with directory content to which we do have READ access to

    1. :param dname: Directory name
    2. :return: List
    """
    dname = posixpath.abspath(dname)
    if not check_directory(dname):
        return []
    ret = []
    for f in os.listdir(dname):
        if not check_file_read("%s/%s"%(dname, f)):
            continue
        ret.append((f, "%s/%s"%(dname, f), os.path.splitext(f)))
    return ret

def read_dir_content(dname, patt="*"):
    """
    Return content of the directory as a list of the string containing the file's content

    1. :param dname: Directory name
    2. :param patt: Pattern matching
    3. :return:
    """
    ret = {}
    for _f, _fn, _fi in get_dir_content(dname):
        if fnmatch.fnmatch(_f, patt) != True:
            continue
        if check_file_read(_fn) != True:
            continue
        ret[_f] = open(_fn).read()
    return ret

def repeat(fun, log_fun, max_attempts, msg="Attempt: "):
    """
    Repeat function X until it returned True, otherwise number of times, with logging

    1. :param fun: Function
    2. :param log_fun: Logging function
    3. :param max_attempts: Maximum number of attempts
    4. :param msg: Add to a message
    5. :return: True/False
    """
    c = 0
    while c < max_attempts:
        log_fun("info", "%s (# %d)"%(msg, c))
        c += 1
        try:
            res = fun()
        except:
            continue
        if res != False:
            return True
    return False

def read_file_into_buffer(fname):
    """
    Reading text file into a list of strings. Empty lines and lines beginning with '#' are ignored

    1. :param fname: Filename
    2. :return: List of strings
    """
    buf = []
    _buf = open(fname).readlines()
    for l in _buf:
        _l = l.strip()
        if len(_l) == 0 or _l[0] == '#':
            continue
        buf.append(_l)
    return buf

def buffer_2_list(_buffer):
    out = []
    for l in _buffer.split("\n"):
        l = l.strip()
        if len(l) == 0 or l[0] == "#":
            continue
        out.append(l)
    return out
def list_2_buffer(_list):
    return "".join(_list)

def get_directory_name(path):
    """
    Get the basename of the directory

    1. :param path: Directory path
    2. :return: basename
    """
    import posixpath
    if check_directory(path) != True:
        return None
    return posixpath.basename(posixpath.abspath(path))

def get_from_env(*var_names, **kw):
    _default = None
    if kw.has_key("default"):
        _default = kw["default"]
    for e in var_names:
        if kw.has_key("kw") and kw["kw"].has_key(e):
            return kw["kw"][e]
        elif os.environ.has_key(e):
            return os.environ[e]
        else:
            pass
    return _default

def compress(data, clibs=[]):
    if len(clibs) == 0:
        return (False, None, data)
    for c in clibs:
        try:
            if c == "zlib":
                import zlib
                return (True, "zlib", zlib.compress(data))
            elif c == "lz4":
                import lz4
                return (True, "zlib", lz4.compress(data))
            elif c == "snappy":
                import pysnappy
                return (True, "snappy", pysnappy.compress(data))
            else:
                import zlib
                return (True, "zlib", zlib.compress(data))
        except:
            return (False, None, data)
    return (False, None, data)

def decompress(data, c=None):
    if c == None:
        return (False, None, data)
    try:
        if c == "zlib":
            import zlib
            return (True, "zlib", zlib.decompress(data))
        elif c == "lz4":
            import lz4
            return (True, "zlib", lz4.decompress(data))
        elif c == "snappy":
            import pysnappy
            return (True, "snappy", pysnappy.uncompress(data))
        else:
            import zlib
            return (True, "zlib", zlib.decompress(data))
    except:
        return (False, None, data)

def split_list(s, sep=":"):
    res = []
    p = s.strip().split(sep)
    for i in p:
        res.append(i.strip())
    return res

def list_rotate(_list, n=1):
    return _list[n:] + _list[:n]

def list2(b, n, l=[]):
    if len(l) == n:
        return l
    if len(l) == 0:
        return list2(b, n, [b, ])
    else:
        l.append(l[-1] * 2)
        return list2(b, n, l)

def banner(s):
    try:
        from pyfiglet import Figlet
    except ImportError:
        return s
    f = Figlet()
    return f.renderText(s)

def dehumanize_time(_str, _default):
    try:
        import humanfriendly
        return humanfriendly.parse_timespan(_str)
    except KeyboardInterrupt:
        return _default


def string_to_quoted_expr(s):
    return HyList(parser.parse(lexer.lex(s)))

def zq_eval(_line, env=None, _shell=None):
    if env != None and not env.cfg["ZQ_UNSAFE_GLOBALS"]:
        if _shell != None:
            _shell.ok("Using safe globals")
        e = env.Globals
    else:
        if _shell != None:
            _shell.warning("Using unsafe globals")
        e = globals()
    return hy_eval(string_to_quoted_expr(_line), e, 'zq')[0]

def load_file_from_the_reference(_ref):
    if _ref[0] == "+":
        ## This is a file path
        _ref = _ref[1:]
        if not check_file_read(_ref):
            return None
        return open(_ref).read()
    elif _ref[0] == "@":
        ## This is URL
        _ref = _ref[1:]
        try:
            url = urllib.FancyURLopener()
            s = url.open(_ref).read().strip()
            url.close()
        except:
            return None
        finally:
            return s
    else:
        ## Return the value as is
        return _ref

def load_and_parse_from_the_reference(_ref, kw):
    buf = load_file_from_the_reference(_ref)
    if not buf:
        return buf
    try:
        tpl = string.Template(buf)
        res = tpl.safe_substitute(kw)
    except:
        return None
    finally:
        return res


def check_reference_read(_ref, _dir=False):
    if _ref[0] == "@":
        ## In 0.2 we do not know how to check URL references
        return True
    if _ref[0] == "+":
        _ref = _ref[1:]
    if _dir:
        return check_directory(_ref)
    else:
        return check_file_read(_ref)

def print_dict(_shell, _dict):
    if _shell == None:
        return
    for i in _dict.keys():
        _shell.ok(": %-25s : %-50s ;"%(i,repr(_dict[i])))

def extract_key_from_info(_info, main_key, search_key):
    res = []
    if not _info.has_key(main_key):
        return res
    if type(_info[main_key]) != types.ListType:
        return res
    for d in _info[main_key]:
        if type(d) == types.DictType and d.has_key(search_key):
            res.append(d[search_key])
    return res

def list2listofdicts(_list, key):
    out = []
    for i in _list:
        out.append({key: i})
    return out

def listofdict2list(_list, _key):
    out = []
    for i in _list:
        if i.has_key(_key):
            out.append(i[_key])
    return out

def discover_element(ctx, _type, _elem):
    if not _elem.has_key(_type):
        ctx.push(_elem)
    else:
        return _elem[_type]
    return None


def do_template(_buffer, _kw=os.environ, **kw):
    for k in kw.keys():
        _kw[k] = kw[k]
    tpl = string.Template(_buffer)
    res = tpl.safe_substitute(_kw)
    return res





def Query(ctx, _name, _reference):
    q = load_file_from_the_reference(_reference)
    if q == None:
        Status(ctx, STATUS=False, MESSAGE="Can not load query %s from %s"%(_name, _reference))
        return ctx
    q =list_2_buffer(buffer_2_list(q))
    ctx.env.Queries[_name] = q
    Status(ctx, STATUS=True, MESSAGE="Query %s loaded from %s" % (_name, _reference))
    return ctx

def Load(ctx, *_name, **kw):
    if kw.has_key("ref"):
        _references = kw["ref"]
    else:
        _references = ctx.env.cfg["ZQ_REF_BASE"]
    for _qn in _name:
        is_loaded = False
        for _ref in _references:
            full_reference = "%s/%s.zql"%(_ref, _qn)
            _ctx = Query(ctx, _qn, full_reference)
            status = ctx.pull()
            status = discover_element(ctx, "STATUS", status)
            if status == None:
                if ctx.env.shell != None:
                    ctx.env.shell.error("Expected STATUS but got something else. Check your logic!")
                continue
            if status["STATUS"] == True:
                is_loaded = True
                break
        if is_loaded == False and ctx.env.shell != None:
            ctx.env.shell.error("Can not load query %s from the references given"%_qn)
        elif is_loaded == True and ctx.env.shell != None:
            ctx.env.shell.ok("(Load %s) returns TRUE" % _qn)
        else:
            pass
    return ctx

def LoadPath(ctx, *path_list, **kw):
    ## Let set some useful vars here
    kw["CWD"] = os.getcwd()
    kw["HOME"] = os.environ["HOME"]
    for p in path_list:
        real_path = posixpath.abspath(do_template(p, kw))
        if not check_directory(real_path):
            if ctx.env.shell != None:
                ctx.env.shell.warning("(LoadPath...) can not access to the directory %s"%real_path)
        if ctx.env.shell != None:
            ctx.env.shell.ok("(LoadPath...) from %s" % real_path)
        for f in os.listdir(real_path):
            real_file = "%s/%s"%(real_path, f)
            if not check_file_read(real_file):
                if ctx.env.shell != None:
                    ctx.env.shell.warning("(LoadPath...) can not access file %s" % real_file)
                continue
            _name = rchop(f, ".zql")
            _ctx = Query(ctx, _name, "+"+real_file)
            if ctx.env.shell != None:
                ctx.env.shell.ok("(LoadPath...) loaded %s from %s " % (_name, real_file))

    return ctx


def Call(ctx, _name, **kw):
    tpl = string.Template(ctx.env.Queries[_name])
    real_query = tpl.safe_substitute(kw)
    _ctx = ctx.env.QUERY(real_query, ctx=ctx)
    return ctx



import time

def color(msg, _f=None, _b=None):
    try:
        import termcolor
    except ImportError:
        return msg
    return termcolor.colored(msg, _f, _b)

def white(msg):
    return color(msg, "white")
def grey(msg):
    return color(msg, "grey")

def _str_msg(msg, _t, _label="UNKNOWN"):
    _t = color("[%s]"%_label, _t)
    _s = color("%s" % time.strftime("%H:%M:%S"), "blue")
    _sep = color(":", "white")
    return "%s %-25s %s %-10s %s %-s" % (_sep, _t, _sep, _s, _sep, color(msg, "cyan"))

def str_error(msg):
    return _str_msg(msg, "red", "ERROR")
def str_warning(msg):
    return _str_msg(msg, "yellow", "WARNING")
def str_ok(msg):
    return _str_msg(msg, "green", "OK")
def str_important(msg):
    return _str_msg(msg, "magenta", "IMPORTANT")


def console_error(msg):
    print str_error(msg)
def console_warning(msg):
    print str_warning(msg)
def console_ok(msg):
    print str_ok(msg)
def console_important(msg):
    print str_important(msg)


def str_dict(d):
    out = ""
    _sep = color(":", "yellow")
    keys = d.keys()
    keys.sort()
    for k in keys:
        out += "%s %-36s %s %-60s\n"%(_sep, color(k,"cyan"), _sep, color(str(d[k]),"white"))
    return out

class ZQ_TERMINAL:
    def __init__(self):
        self.parser.add_argument("-v",  action="count", help="Increase verbosity")
    def ok(self, msg, **kw):
        if self.log != None:
            self.log.ok(msg, kw)
        if self.args.v == None or self.args.v < 2:
            return
        console_ok(msg%kw)
    def warning(self, msg, **kw):
        if self.log != None:
            self.log.ok(msg, kw)
        if self.args.v == None or self.args.v < 1:
            return
        console_warning(msg % kw)
    def error(self, msg, **kw):
        if self.log != None:
            self.log.ok(msg, kw)
        console_error(msg % kw)
    def important(self, msg, **kw):
        console_important(msg % kw)




def Error(ctx, msg):
    if ctx.env.shell != None:
        ctx.env.shell.error(msg)
    return ctx

def Warning(ctx, msg):
    if ctx.env.shell != None:
        ctx.env.shell.warning(msg)
    return ctx

def Ok(ctx, msg):
    if ctx.env.shell != None:
        ctx.env.shell.ok(msg)
    return ctx

def RawDisplay(ctx, *data, **kw):
    if kw.has_key("repr") and kw["repr"] == True:
        for i in data:
            print repr(i),
    else:
        for i in data:
            print i,
    return ctx


def _create_hostgroup(ctx, _data):
    try:
        res = apply(ctx.zapi.hostgroup.create, (), _data)
    except:
        return False
    if not res.has_key("groupids"):
        return False
    return True

def _create_host(ctx, _data):
    print "HHH",_data
    return True

_CREATE_CALL_TABLE={
    "HOSTGROUPS": _create_hostgroup,
    "HOST": _create_host,

}

def Create(ctx, *args, **kw):
    if not _fill_bjq_queue(ctx, ctx.jobs.direct, "CREATE"):
        if ctx.env.shell != None:
            ctx.env.shell.warning("(Create...) can not populate the queue. Check logic!")
        return ctx
    while True:
        sjob = ctx.jobs.pull()
        if sjob == None:
            break
        _pri, _req = sjob
        _key, _data = _req
        if not _CREATE_CALL_TABLE[_key](ctx, _data):
            if ctx.env.shell != None:
                ctx.env.shell.warning("(Create...) can not create element %s"%str(_req))
    return ctx





def Version(ctx):
    ctx.push({'VERSION': ctx.zapi.api_version()})
    return ctx

def Hosts(ctx, **kw):
    try:
        res = apply(ctx.zapi.host.get, (), kw)
        ctx.push({'HOST': res})
    except:
        if ctx.env.shell != None:
            ctx.env.shell.error("Error in submitting (Hosts) query to Zabbix")
        return ctx
    return ctx

def Interfaces(ctx):
    ctx.push({'INTERFACES': ctx.zapi.hostinterface.get()})
    return ctx

def Hostgroups(ctx, **kw):
    try:
        res = apply(ctx.zapi.hostgroup.get, (), kw)
        ctx.push({'HOSTGROUPS': res})
    except:
        if ctx.env.shell != None:
            ctx.env.shell.error("Error in submitting (Hostgroups) query to Zabbix")
        return ctx
    return ctx


def Templates(ctx):
    ctx.push({'TEMPLATES': ctx.zapi.template.get()})
    return ctx

def Items(ctx):
    ctx.push({'ITEMS': ctx.zapi.item.get()})
    return ctx

def Triggers(ctx):
    ctx.push({'TRIGGERS': ctx.zapi.trigger.get()})
    return ctx

def Proxies(ctx):
    ctx.push({'PROXIES': ctx.zapi.proxy.get()})
    return ctx




def IfTrue(ctx, *query_names, **kw):
    stat = ctx.pull()
    if not stat:
        return ctx
    if stat != True and not stat.has_key("STATUS") and stat["STATUS"] != True :
        ctx.push(stat)
        return ctx
    for i in stat["STATUS"].keys():
        kw[i] = stat["STATUS"][i]
    for query_name in query_names:
        _ctx = apply(Call, (ctx, query_name), kw)
    return ctx

def IfFalse(ctx, *query_names, **kw):
    stat = ctx.pull()
    if not stat:
        return ctx
    if stat != False and not stat.has_key("STATUS") and stat["STATUS"] != False:
        ctx.push(stat)
        return ctx
    for i in stat["STATUS"].keys():
        kw[i] = stat["STATUS"][i]
    for query_name in query_names:
        _ctx = apply(Call, (ctx, query_name ), kw)
    return ctx

def Until(ctx, key, *query_names, **kw):
    while True:
        data = ctx.pull()
        if data == None:
            return ctx
        if type(data) != type({}) or not data.has_key(key):
            ctx.push(data)
            return ctx
        kw["DATA"] = data[key]
        for query_name in query_names:
            _ctx = apply(Call, (ctx, query_name), kw)
    ## We should never reach this point, but just in case
    return ctx



include "zq.pyx"
include "zq_terminal.pyx"
include "zq_query.pyx"
include "zq_cmd.pyx"
include "zq_cmd_exec.pyx"
include "zq_cmd_eval.pyx"
include "zq_cmd_query.pyx"

class ZQ_SHELL(ZQ_GEN,ZQ_TERMINAL,ZQ_HELP,ZQ_CMD_EVAL,ZQ_CMD_QUERY,ZQ_CMD_EXEC):
    def __init__(self):
        self.doc = []
        self.log = None
        ZQ_GEN.__init__(self, "zql v %s" % ZQ_VERSION, "zql - Zabbix Query Shell")
        ZQ_TERMINAL.__init__(self)
        ZQ_HELP.__init__(self)
        ZQ_CMD_EVAL.__init__(self)
        ZQ_CMD_QUERY.__init__(self)
        self.BANNER = "ZQL %s"%ZQ_VERSION
    def main_preflight(self):
        self._main_preflight()
    def process(self):
        if ZQ_GEN.process(self) != True:
            self.error("There is an error and I can nor recover from it. Exit")
            sys.exit(99)
        return True

def main():
    cmd = ZQ_SHELL()
    cmd.process()

main()


Eq = lambda x,y: x == y
Ne = lambda x,y: x != y
Gt = lambda x,y: x > y
Gte = lambda x,y: x >= y
Lg = lambda x,y: x < y
Lge = lambda x,y: x <= y

def Match(x,y):
    import re,fnmatch
    if fnmatch.fnmatch(x,y) == True:
        return True
    else:
        return False

TRUE = lambda x,y: True
FALSE = lambda x,y: False
T = 1
F = 0
NONE = None
NEW = "NEW"
PULL = "PULL"
DEFAULT = "DEFAULT"
CR = "\n"



hostid="hostid"
hostids="hostids"
groupid="groupid"
groupids="groupids"
templateid="templateid"
templateids="templateids"
triggerid="triggerid"
triggerids="triggerids"

status="status"
name="name"
proxyid="proxyid"
proxy_hostid="proxy_hostid"

## Action codes and symbols
limit="limit"
filter="filter"
extend="extend"
GET=0
CREATE=1
DELETE=2
UPDATE=3

## Symbols for the (New...)
HOST="HOST"
HOSTGROUPS="HOSTGROUPS"




import traceback

class ZQ_CMD_EVAL:
    def __init__(self):
        self.parser.add_argument("--traceback", action="store_true",
                                 help="Display error traceback for the LISP evaluations")
        self.parser.add_argument("--raw", action="store_true",
                                 help="Output result 'as-is', without extra formatting")
        self.parser.add_argument("--no-out", action="store_true",
                                 help="Supress the output of the query result")
        self.parser.add_argument("--default-environment", type=str, default="default",
                                 help="Set the default environment name")
        self.parser.add_argument("--default-server", type=str,
                                 help="Set the default server for the queries")
        self.parser.add_argument("--unsafe-globals", action="store_true",
                                 help="Use unsafe globals")

    def preflight(self):
        self.env.cfg["ZQ_UNSAFE_GLOBALS"] = self.args.unsafe_globals
    def make_doc(self):
        self.doc.append(("eval", "evaluate LISP statements"))
    def HELP_EVAL(self):
        print """Evaluate LISP functions:
"eval" "(lisp functionm)" "(lisp function)"
Example:
    %s eval "(time.time)"
                """%sys.argv[0]
    def EVAL(self):
        self.Exec(self.env.EVAL)



def load_config_file(path):
    if path[0] != '@' and path[0] != "+":
        # Conside this to be a file
        path = "+" + path
    cfg = load_file_from_the_reference(path)
    if cfg == None:
        return None
    cfg_type = path.split(".")[-1].lower()
    if cfg_type == "json":
        import simplejson
        try:
            return simplejson.loads(cfg)
        except KeyboardInterrupt:
            return None
    elif cfg_type == "ini":
        import ConfigParser
        import StringIO
        try:
            config = ConfigParser.SafeConfigParser()
            buf = StringIO.StringIO(cfg)
            config.readfp(buf)
            out = []
            for s in config.sections():
                d = {}
                d["name"] = s
                d["url"] = config.get(s, "url")
                d["username"] = config.get(s, "username")
                d["password"] = config.get(s, "password")
                out.append(d)
            return out
        except KeyboardInterrupt:
            return None



class CFG(UserDict.UserDict):
    def __init__(self, d={}, **kw):
        UserDict.UserDict.__init__(self)
        if len(d.keys()) > 0:
            _d = d
        else:
            _d = kw
        for k in _d.keys():
            self[k.upper()] = get_from_env(k.upper(), kw=_d)
    def add_missing(self, name, params, default):
        self[name.upper()] = get_from_env(name.upper(), kw=params, default=default)




class ZQ_ENV(UserDict.UserDict):
    def __init__(self, _shell=None, **kw):
        UserDict.UserDict.__init__(self)
        self.ready = False
        self.shell = _shell
        self.params = kw
        self.Globals = {}
        self.Queries = {}
        self.RefBase = "+."
        self.envs = None
        self.srv = ZQ_SRV(self, self.shell)
        self.ready = self.reload()
    def registerGlobals(self, name, _ref):
        self.Globals[name] = _ref
    def reload(self):
        self.cfg = CFG(self.params)
        self.set_defaults()
        return True
    def set_defaults(self):
        self.cfg["ZQ_HOME"] = "/tmp"
        self.cfg["ZQ_REF_BASE"] = self.RefBase
        self.cfg["ZQ_ENV_NAME"] = "default"
        self.cfg["ZQ_MAX_PIPELINE"] = 100
        self.cfg["ZQ_MAX_ENV_STACK"] = 100
        self.cfg["ZQ_UNSAFE_GLOBALS"] = False
        self.envs = LifoQueue(self.cfg["ZQ_MAX_ENV_STACK"])
        self.registerGlobals("Set", Set)
        self.registerGlobals("Append", Append)
        self.registerGlobals("Ne", Ne)
        self.registerGlobals("Eq", Eq)
        self.registerGlobals("Lg", Lg)
        self.registerGlobals("Lge", Lge)
        self.registerGlobals("Gt", Gt)
        self.registerGlobals("Gte", Gte)
        self.registerGlobals("TRUE", TRUE)
        self.registerGlobals("FALSE", FALSE)
        self.registerGlobals("NONE", NONE)
        self.registerGlobals("NEW", NEW)
        self.registerGlobals("DEFAULT", DEFAULT)
        self.registerGlobals("T", T)
        self.registerGlobals("F", F)
        self.registerGlobals("Match", Match)
        self.registerGlobals("CR", CR)
        ##
        self.registerGlobals("hostid", hostid)
        self.registerGlobals("hostids", hostids)
        self.registerGlobals("templateid", templateid)
        self.registerGlobals("templateids", templateids)
        self.registerGlobals("groupid", groupid)
        self.registerGlobals("groupids", groupids)
        self.registerGlobals("status", status)
        self.registerGlobals("proxyid", proxyid)
        self.registerGlobals("proxy_hostid", proxy_hostid)
        ##
        self.registerGlobals("ZBX", ZBX)
        self.registerGlobals("Filter", Filter)
        self.registerGlobals("Hosts", Hosts)
        self.registerGlobals("Templates", Templates)
        self.registerGlobals("Items", Items)
        self.registerGlobals("Proxies", Proxies)
        self.registerGlobals("Triggers", Triggers)
        self.registerGlobals("Hostgroups", Hostgroups)
        self.registerGlobals("Interfaces", Interfaces)
        self.registerGlobals("Version", Version)
        self.registerGlobals("Value", Value)
        self.registerGlobals("Drop", Drop)
        self.registerGlobals("Push", Push)
        self.registerGlobals("New", New)
        self.registerGlobals("Peek", Peek)
        self.registerGlobals("Until", Until)
        self.registerGlobals("Swap", Swap)
        self.registerGlobals("Setv", Setv)
        self.registerGlobals("Getv", Getv)
        self.registerGlobals("Dup", Dup)
        self.registerGlobals("Merge", Merge)
        self.registerGlobals("Out", Out)
        self.registerGlobals("File", File)
        self.registerGlobals("Json", Json)
        self.registerGlobals("Pretty_Json", Pretty_Json)
        self.registerGlobals("Status", Status)
        self.registerGlobals("Result", Result)
        self.registerGlobals("Group", Group)
        self.registerGlobals("Ungroup", Ungroup)
        self.registerGlobals("Delete", Delete)
        self.registerGlobals("Create", Create)
        self.registerGlobals("Update", Update)
        self.registerGlobals("Query", Query)
        self.registerGlobals("Call", Call)
        self.registerGlobals("Load", Load)
        self.registerGlobals("LoadPath", LoadPath)
        self.registerGlobals("IfTrue", IfTrue)
        self.registerGlobals("IfFalse", IfFalse)
        self.registerGlobals("Error", Error)
        self.registerGlobals("Warning", Warning)
        self.registerGlobals("Ok", Ok)
        self.registerGlobals("RawDisplay", RawDisplay)
        self.registerGlobals("Clear", Clear)
        self.registerGlobals("ClearJobQueue", ClearJobQueue)
        ##
        self.registerGlobals("GET", GET)
        self.registerGlobals("CREATE", CREATE)
        self.registerGlobals("DELETE", DELETE)
        self.registerGlobals("UPDATE", UPDATE)
        self.registerGlobals("limit", limit)
        self.registerGlobals("filter", filter)
        ##
        self.registerGlobals("HOST", HOST)
        self.registerGlobals("HOSTGROUPS", HOSTGROUPS)

    def EVAL(self, _q):
        return zq_eval(_q, self, self.shell)
    def QUERY(self, _q, default_environment="default", default_server=None, **kw):
        env = ENVIRONMENT(default_environment)
        if default_server != None and not kw.has_key("ctx"):
            query = '(-> (ZBX NONE "%s" "%s") %s)'%(default_server, default_environment, _q)
        elif default_server == None and not kw.has_key("ctx"):
            query = "(-> %s)" % _q
        elif kw.has_key("ctx"):
            kw["ctx"].ctx_push(kw["ctx"])
            query = '(-> (ZBX DEFAULT "" "%s") %s)'%(default_environment, _q)
        else:
            query = "(-> %s)" % _q
        if self.shell != None:
            self.shell.ok("Translated query: %s"%query)
        res = zq_eval(query, self, self.shell)
        return res

class ENV_CTL(UserDict.UserDict):
    def __init__(self):
        UserDict.UserDict.__init__(self)
        self.current = None
    def set_current(self, name):
        if name in self.keys():
            self.current = name
        return self[name]
    def __setitem__(self, key, val):
        self.current = key
        return UserDict.UserDict.__setitem__(self, key, val)
    def __delitem__(self, key):
        UserDict.UserDict.__delitem__(self, key)
        if self.current == key:
            self.current = None

ENV = ENV_CTL()

def ENVIRONMENT(name, _env=None):
    global ENV
    if ENV.has_key(name):
        return ENV[name]
    elif ENV.has_key("default"):
        ENV["default"] = _env
    elif _env != None:
        ENV[name] = _env
    else:
        ENV[name] = ZQ_ENV()
    return ENV[name]



def Out(ctx, key = None):
    data = ctx.pull()
    if not data:
        return ""
    if key and type(data) == type({}) and data.has_key(key):
        return data[key]
    if type(data) in [type(""), type(u"")]:
        return str(data)
    return data

def File(data, filename):
    try:
        f = open(filename, "w")
        f.write(str(data))
        f.close()
    except:
        return ""
    return filename

def Msgpack(data):
    import msgpack
    return msgpack.dumps(data)

def Compress(data):
    import zlib
    return zlib.compress(data)

def Pretty_Json(data):
    if type(data) == type({}):
        return simplejson.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))
    return data

def Json(data):
    if type(data) == type({}):
        return simplejson.dumps(data)
    return data




def F(ctx, *_arg_filter, **_kw_filter):
    """
    This word will push filters in the stack
    :param ctx:
    :param _arg_filter:
    :param _kw_filter:
    :return:
    """
    for _f in _kw_filter.keys():
        ctx.push({'FILTER': {'TYPE':_f, 'FILTER':_kw_filter[_f]}})
    for _key, _f  in _arg_filter:
        ctx.push({'FILTER': {'TYPE':_key, 'FILTER': _f}})
    return ctx

def Dup(ctx):
    import copy
    p1 = ctx.pull()
    if not p1:
        return ctx
    p2 = copy.deepcopy(p1)
    ctx.push(p2)
    ctx.push(p1)
    return ctx
def Swap(ctx):
    p1 = ctx.pull()
    if not p1:
        return ctx
    p2 = ctx.pull()
    if not p2:
        ctx.push(p1)
        return ctx
    ctx.push(p1)
    ctx.push(p2)
    return ctx

def Drop(ctx):
    ctx.pull()
    return ctx

def Push(ctx, key, value=None, **kw):
    if value == None:
        ctx.push({key: kw})
    else:
        ctx.push({key:value})
    return ctx

def Peek(ctx):
    data = ctx.pull()
    if data:
        ctx.push(data)
    print data
    return ctx


def Merge(ctx, dst_merge_key, merge_key=None):
    merging_data = ctx.pull()
    if merging_data == None:
        return ctx
    merging_with = ctx.pull()
    if merging_with == None:
        ctx.push(merging_data)
        return ctx
    if dst_merge_key == None:
        dst_merge_key = merge_key
    new_data = {}
    for dst_key in merging_with.keys():
        new_data[dst_key] = []
        for dst_vals in merging_with[dst_key]:
            for src_key in merging_data.keys():
                dst_vals[src_key] = []
                for src_vals in merging_data[src_key]:
                    if (src_vals.has_key(merge_key) and dst_vals.has_key(dst_merge_key)) and (src_vals[merge_key] == dst_vals[dst_merge_key]):
                        dst_vals[src_key].append(src_vals)
            new_data[dst_key].append(dst_vals)
    ctx.push(new_data)
    return ctx







def _delete_hostgroup(ctx, args, _data):
    _groupids = listofdict2list(_data, "groupid")
    if len(_groupids) == 0:
        return True
    ret = apply(ctx.zapi.hostgroup.delete, _groupids, {})
    if ret.has_key("groupids"):
        return True
    return False

def _delete_host(ctx, args, _data):
    print "(Delete) HOST",_data
    return True

_DELETE_CALL_TABLE={
    "HOSTGROUPS": _delete_hostgroup,
    "HOST": _delete_host,
}

def Delete(ctx, *args, **kw):
    if kw.has_key("all") and kw["all"] == True:
        limit = None
        if ctx.env.shell != None:
            ctx.env.shell.warning(
                "You had requested to (Delete) all configuration entities that you have in your stack. Hope you know what you are doing.")
    else:
        limit = 1
    if not _fill_bjq_queue(ctx, ctx.jobs.reverse, limit=limit):
        if ctx.env.shell != None:
            ctx.env.shell.warning("(Delete...) can not populate the queue. Check logic!")
        return ctx
    while True:
        sjob = ctx.jobs.pull()
        if sjob == None:
            break
        _pri, _req = sjob
        _key, _data = _req
        if not _DELETE_CALL_TABLE[_key](ctx, args, _data):
            if ctx.env.shell != None:
                ctx.env.shell.warning("(Delete...) can not delete element %s"%str(_req))
    return ctx



from setuptools import setup#, find_packages, Extension
import distutils.command.build as _build
import setuptools.command.install as _install

import sys
import os
import os.path as op
import distutils.spawn as ds
import distutils.dir_util as dd
import posixpath


def run_cmake(arg=""):
    """
    Forcing to run cmake
    """
    if ds.find_executable('cmake') is None:
        print "CMake  is required to build zql"
        print "Please install cmake version >= 2.8 and re-run setup"
        sys.exit(-1)

    print "Configuring zql build with CMake.... "
    cmake_args = arg
    try:
        build_dir = op.join(op.split(__file__)[0], 'build')
        dd.mkpath(build_dir)
        os.chdir("build")
        ds.spawn(['cmake', '..'] + cmake_args.split())
        ds.spawn(['make', 'clean'])
        ds.spawn(['make'])
        os.chdir("..")
    except ds.DistutilsExecError:
        print "Error while running cmake"
        print "run 'setup.py build --help' for build options"
        print "You may also try editing the settings in CMakeLists.txt file and re-running setup"
        sys.exit(-1)

class build(_build.build):
    def run(self):
        run_cmake()
        # Now populate the extension module  attribute.
        #self.distribution.ext_modules = get_ext_modules()
        _build.build.run(self)

class install(_install.install):
    def run(self):
        if not posixpath.exists("src/zq.so"):
            run_cmake()
        ds.spawn(['make', 'install'])
        #self.distribution.ext_modules = get_ext_modules()
        self.do_egg_install()

with open('README.txt') as file:
    clips6_long_desc = file.read()

setup(
    name = "zq",
    version = '0.3',
    description = 'ZQL - Zabbix Query Language',
    install_requires = ["cython", "msgpack-python", "simplejson", "hy", "pyfiglet",
                "gevent", "json", "termcolor", "humanfriendly"],
    requires = [],
    include_package_data = True,
    url = 'https://github.com/vulogov/zq/',
    author='Vladimir Ulogov',
    author_email = 'vladimir.ulogov@me.com',
    maintainer_email = 'vladimir.ulogov@me.com',
    license = "GNU GPL Versin 3",
    long_description = clips6_long_desc,
    keywords = "zql, monitoring, zabbix",
    platforms = ['GNU/Linux','Unix','Mac OS-X'],
    classifiers = [
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Intended Audience :: Information Technology',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Topic :: Software Development',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: System :: Monitoring',
            'Topic :: System :: Networking :: Monitoring',
            'Topic :: System :: Systems Administration',
            'Environment :: Console',
            'Environment :: Console :: Curses'
    ],
    # ext_modules is not present here. This will be generated through CMake via the
    # build or install commands
    cmdclass={'install':install,'build': build},
    zip_safe=False,
    packages = ['zq'],
    package_data = {
        'zq': ['zq.so', '*.pyx', '*.pyi']
    }
)



##
## Balanced Job Queue
##

def _fill_bjq_queue(ctx, _cmd, _key=None, **kw):
    limit = None
    if kw.has_key("limit") and kw["limit"] != None:
        try:
            limit = int(kw["limit"])
        except:
            limit = None
    c = 0
    while True:
        if limit != None and c >= limit:
            break
        job = ctx.pull()
        if not job:
            break
        if type(job) != types.DictType:
            ## This is not a real job, eih !
            ctx.push(job)
            break
        c += 1
        acceptable = False
        for k in job.keys():
            if not ctx.jobs.isAcceptable(k) or type(job[k]) != types.DictType:
                acceptable = False
            else:
                acceptable = True
        if not acceptable:
            ctx.push(job)
            break
        for k in job.keys():
            if _key != None and k == _key:
                continue
            if _key != None:
                del job[k][_key]
        res = _cmd(job)
        if res == False:
            return False
    return True

class BJQ:
    def __init__(self, env):
        self.supported_types = ["HOSTGROUPS", "TEMPLATES", "HOST", "ITEMS", "ACTIONS"]
        self.env = env
        self.jobs = Queue.PriorityQueue(maxsize=self.env.cfg["ZQ_MAX_PIPELINE"])
    def isAcceptable(self, key):
        return key in self.supported_types
    def straight_order(self):
        return self.supported_types
    def reverse_order(self):
        return self.supported_types[::-1]
    def push(self, _element, _order):
        if type(_element) != types.DictType:
            return False
        for k in _element.keys():
            if k not in _order:
                continue
            pri = _order.index(k)
            self.jobs.put_nowait((pri, (k,_element[k])))
        return True
    def direct(self, _element):
        return self.push(_element, self.straight_order())
    def reverse(self, _element):
        return self.push(_element, self.reverse_order())
    def pull(self):
        try:
            res = self.jobs.get_nowait()
            return res
        except:
            return None



def ClearJobQueue(ctx):
    while True:
        j = ctx.jobs.pull()
        if not j:
            break
    return ctx




def Setv(ctx, name, val):
    if not ctx.env.cfg["ZQ_UNSAFE_GLOBALS"]:
        ctx.env.registerGlobals(name, val)
    return ctx

def Getv(ctx, name):
    if ctx.env.cfg["ZQ_UNSAFE_GLOBALS"]:
        return None
    try:
        return ctx.env.Globals[name]
    except KeyboardInterrupt:
        return None


class ZQ_CMD_EXEC:
    def ExecSingle(self, _callable, s, args):
        try:
            if s[0] == '@':
                ## Hey, it is URL
                url = urllib.FancyURLopener()
                s = url.open(s[1:]).read().strip()
            elif s[0] == '+':
                ## Hey, it is a file path
                if not check_file_read(s[1:]):
                    return
                s = open(s[1:]).read()
            s = list_2_buffer(buffer_2_list(s))
            a = tuple([s,]+list(args))
            if self.env.shell != None:
                self.env.shell.ok("%s: %s"%(_callable.__name__, s))
            res = apply(_callable, a, {})
            if self.args.v != None or not self.args.raw:
                prefix = color("%s"%s,"cyan")+color(" = ","yellow")
            else:
                prefix = ""
            if self.args.no_out:
                return
            elif not self.args.raw:
                print "%s%s"%(prefix,color(repr(res),"white"))
            else:
                print res
        except:
            self.error("Error in %s"%s)
            if self.args.traceback:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                traceback.print_exception(exc_type, exc_value, exc_traceback,
                                          limit=6, file=sys.stdout)
    def Exec(self, _callable, *args):
        for s in self.args.N[1:]:
            self.ExecSingle(_callable, s, args)



def Status(ctx, *args, **kw):
    if not ctx:
        return ctx
    status = {'TIME': time.time(), 'ID':str(uuid.uuid4())}
    status['DATA'] = args
    for i in kw.keys():
        status[i] = kw[i]
    ctx.push({'STATUS':status})
    return ctx

def Result(ctx, **kw):
    status = ctx.pull()
    if not status:
        return ctx
    if not status.has_key("STATUS"):
        ctx.push(status)
        return ctx
    if status["STATUS"].has_key("MESSAGE"):
        msg = status["STATUS"]["MESSAGE"]%status["STATUS"]
    else:
        msg = repr(status["STATUS"])
    if status["STATUS"].has_key("STATUS") and type(status["STATUS"]["STATUS"]) == types.BooleanType:
        stat = status["STATUS"]["STATUS"]
    else:
        stat = False
    if stat == False and ctx.env.shell != None:
        ctx.env.shell.error(msg)
        ctx.push(False)
    elif stat == True and ctx.env.shell != None:
        ctx.env.shell.ok(msg)
        ctx.push(True)
    else:
        pass
    if kw.has_key("die") and kw["die"] == True:
        sys.exit(0)
    return ctx


from Queue import LifoQueue

class ZQ_ZBX(object):
    def __init__(self, _url, _username, _password, _name, _sender, _sender_port, env, _shell=None):
        self.env = env
        self.shell = _shell
        self.name = _name
        self.url = _url
        self._username = _username
        self._password = _password
        self.sender = _sender
        self.sender_port = _sender_port
        self.zapi = None
        self.value = LifoQueue(self.env.cfg["ZQ_MAX_PIPELINE"])
        self.jobs = BJQ(self.env)
        self.reconnect()
    def reload(self):
        """
        reload() will clear all (but jobs) context queues and reconnect to Zabbix
        :return:
        """
        self.reconnect()
        while True:
            p = self.pull()
            if not p:
                break
        while True:
            p = self.ctx_pull()
            if not p:
                break
    def reconnect(self):
        try:
            self.zapi = ZabbixAPI(server=self.url)
            self.zapi.login(user=self._username, password=self._password)
        except:
            self.zapi = None
            if self.shell != None:
                self.shell.error("Can not connect to the %s"%self.url)
    def pull(self):
        try:
            return self.value.get_nowait()
        except:
            return None
    def peek(self):
        data = self.pull()
        self.push(data)
        return data
    def push(self, data):
        try:
            self.value.put(data)
        except KeyboardInterrupt:
            return False
        return True
    def ctx_pull(self):
        try:
            return self.env.envs.get_nowait()
        except:
            return None
    def ctx_push(self, _env):
        try:
            self.env.envs.put(_env)
        except:
            return False
        return True


class ZQ_SRV(UserDict.UserDict):
    def __init__(self, env=None, _shell=None, **kw):
        UserDict.UserDict.__init__(self)
        self.env = env
        self.shell = _shell
        self.ready = False
        self.params = kw
    def addServer(self, _url, _username, _password, _name, _sender, _sender_port):
        if not _name:
            _key = _url
        else:
            _key = _name
        if _key not in self.keys():
            self[_key] = ZQ_ZBX(_url,_username,_password,_name,_sender, _sender_port, self.env,self.shell)
        return self[_key]

def ZBX(ctx=None, server="zabbix", name="default"):
    env = ENVIRONMENT(name)
    if ctx not in ["PULL", "DEFAULT"] and not env.srv.has_key(server):
        return None
    if ctx == "PULL":
        return env.envs.get_nowait()
    elif ctx == "DEFAULT":
        _ctx = env.envs.get_nowait()
        if _ctx == None and env.shell != None:
            env.shell.error("DEFAULT context has been requested, but there is None. Check logic!")
            return None
        return _ctx
    else:
        pass
    return env.srv[server]
def ZBXS(name="default"):
    env = ENVIRONMENT(name)
    return env.srv.data





